from __future__ import annotations

import os
from pathlib import Path
from subprocess import Popen
from typing import Optional

from bmde.core import logging
from bmde.core.exec import ExecOptions
from .service import GitService
from .settings import GitSettings
from .spec import GitSpec, GitSpecOpts
from ...config.loader import load_settings
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import BackendOptions

log = logging.get_logger(__name__)


def create_git_spec(
    d: Optional[Path],
    arguments: Optional[list[str]] = None,
    backend: Optional[BackendOptions] = None,
    interactive: bool = True,
    dry_run: bool = True,
    entrypoint: Optional[Path] = None,
    ssh_username: Optional[str] = None,
    ssh_password: Optional[str] = None,
    ssh_host: Optional[str] = None,
    git_name: Optional[str] = None,
    git_email: Optional[str] = None,
    vpn_username: Optional[str] = None,
    vpn_password: Optional[str] = None,
    vpn_host: Optional[str] = None,
    vpn_port: Optional[int] = None,
    vpn_realm: Optional[str] = None,
    vpn_cert: Optional[str] = None,
    vpn_test_dns: Optional[str] = None,
    vpn_test_ip: Optional[str] = None,
    settings: Optional[GitSettings] = None,
) -> GitSpec:
    if d is None:
        d = Path(os.getcwd())

    if settings is None:
        settings = GitSettings()

    return GitSpec(
        GitSpecOpts=GitSpecOpts(
            d=d,
            ssh_username=(
                ssh_username if ssh_username is not None else settings.ssh.username
            ),
            ssh_password=(
                ssh_password if ssh_password is not None else settings.ssh.password
            ),
            ssh_host=ssh_host if ssh_host is not None else settings.ssh.host,
            git_name=git_name if git_name is not None else settings.git.name,
            git_email=git_email if git_email is not None else settings.git.email,
            vpn_username=(
                vpn_username if vpn_username is not None else settings.vpn.username
            ),
            vpn_password=(
                vpn_password if vpn_password is not None else settings.vpn.password
            ),
            vpn_host=vpn_host if vpn_host is not None else settings.vpn.host,
            vpn_port=vpn_port if vpn_port is not None else settings.vpn.port,
            vpn_realm=vpn_realm if vpn_realm is not None else settings.vpn.realm,
            vpn_cert=vpn_cert if vpn_cert is not None else settings.vpn.cert,
            vpn_test_dns=(
                vpn_test_dns if vpn_test_dns is not None else settings.vpn.test_dns
            ),
            vpn_test_ip=(
                vpn_test_ip if vpn_test_ip is not None else settings.vpn.test_ip
            ),
        ),
        SpecExecOpts=SpecExecOpts(
            backend=(
                backend
                if backend is not None
                else (
                    settings.execution_settings.backend
                    if settings.execution_settings.backend is not None
                    else BackendOptions.DOCKER
                )
            ),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=arguments,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                settings.execution_settings.background
                if settings.execution_settings.background is not None
                else False
            ),
            interactive=(
                interactive
                if interactive is not None
                else settings.execution_settings.interactive
            ),
        ),
    )


def execute_git(spec: GitSpec) -> int | Popen[bytes]:
    return GitService().run(
        spec.GitSpecOpts,
        ExecOptions(
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            interactive=spec.SpecExecOpts.interactive,
            entrypoint=spec.SpecExecOpts.entrypoint,
            arguments=spec.SpecExecOpts.arguments,
            backend=spec.SpecExecOpts.backend,
        ),
    )


def git_command(
    d: Optional[Path],
    arguments: Optional[list[str]] = None,
    backend: Optional[BackendOptions] = None,
    dry_run: bool = False,
    interactive: bool = True,
    entrypoint: Optional[Path] = None,
    ssh_username: Optional[str] = None,
    ssh_password: Optional[str] = None,
    ssh_host: Optional[str] = None,
    git_name: Optional[str] = None,
    git_email: Optional[str] = None,
    vpn_username: Optional[str] = None,
    vpn_password: Optional[str] = None,
    vpn_host: Optional[str] = None,
    vpn_port: Optional[int] = None,
    vpn_realm: Optional[str] = None,
    vpn_cert: Optional[str] = None,
    vpn_test_dns: Optional[str] = None,
    vpn_test_ip: Optional[str] = None,
    settings: Optional[GitSettings] = None,
) -> int | Popen[bytes]:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(settings=full_settings)
        settings = full_settings.git

    spec = create_git_spec(
        d=d,
        arguments=arguments,
        backend=backend,
        dry_run=dry_run,
        interactive=interactive,
        entrypoint=entrypoint,
        ssh_username=ssh_username,
        ssh_password=ssh_password,
        ssh_host=ssh_host,
        git_name=git_name,
        git_email=git_email,
        vpn_username=vpn_username,
        vpn_password=vpn_password,
        vpn_host=vpn_host,
        vpn_port=vpn_port,
        vpn_realm=vpn_realm,
        vpn_cert=vpn_cert,
        vpn_test_dns=vpn_test_dns,
        vpn_test_ip=vpn_test_ip,
        settings=settings,
    )

    return execute_git(spec)



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from bmde.core.spec import BaseSpec
from bmde.commands.run.spec import RunSpec
from bmde.core.spec_opts import SpecExecOpts
from bmde.core.types import DockerOutputOptions


@dataclass
class DebugSpecOpts(BaseSpec):
    elf: Path
    docker_screen: Optional[DockerOutputOptions]
    docker_network: Optional[str]
    RunSpec: RunSpec


@dataclass
class DebugSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    DebugSpecOpts: DebugSpecOpts



from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from bmde.config.command_settings import ExecutionSettings
from bmde.core.types import DockerOutputOptions
from bmde.core.types import DOCKER_DESMUME_DEBUG_NETWORK


class RunSettings(BaseModel):
    graphical_output: Optional[DockerOutputOptions] = DockerOutputOptions.HOST

    execution_settings: ExecutionSettings = ExecutionSettings()

    debug: Optional[bool] = False
    arm9_debug_port: Optional[int] = 1024
    fat_image: Optional[Path] = None
    docker_network: Optional[str] = DOCKER_DESMUME_DEBUG_NETWORK



import shutil
import subprocess

from bmde.core.exec import run_cmd, ExecOptions
from .backend import RunBackend
from ..spec import RunSpecOpts


class FlatpakRunner(RunBackend):
    def is_available(self) -> bool:
        if shutil.which("flatpak") is not None:
            return True
        return False

    def run(
        self, spec: RunSpecOpts, exec_opts: ExecOptions
    ) -> int | subprocess.Popen[bytes]:
        # Adjust to your flatpak package and args
        args = ["flatpak", "run", "org.desmume.DesmuME", str(spec.nds_rom)]
        return run_cmd(args, exec_opts)



import os
from pathlib import Path

import typer

from bmde.commands.build.command import build_command
from bmde.config.schema import Settings
from bmde.core import logging
from bmde.core.shared_options import ArgumentsOpt, DirectoryOpt, BackendOpt, EntrypointOpt, DryRunOpt, BackgroundOpt, InteractiveOpt

log = logging.get_logger(__name__)

def build_controller(
        ctx: typer.Context,
        directory: DirectoryOpt = Path(os.getcwd()),
        arguments: ArgumentsOpt = None,
        backend: BackendOpt = None,
        background: BackgroundOpt = False,
        interactive: InteractiveOpt = True,
        entrypoint: EntrypointOpt = None,
        dry_run: DryRunOpt = False
) -> None:
    """devkitARM make wrapper. Builds NDS ROM from source code."""
    settings: Settings = ctx.obj["settings"]

    log.debug("CLI options provided:\n"
              f"- Arguments: {str(arguments)}\n"
              f"- Directory: {str(directory)}\n"
              f"- Backend: {str(backend)}\n"
              f"- Background: {str(background)}\n"
              f"- Interactive: {str(interactive)}\n"
              f"- Entrypoint: {str(entrypoint)}\n"
              f"- Dry run: {str(dry_run)}\n"
              )

    ret = build_command(
        d=directory,
        arguments=arguments,
        backend=backend,
        background=background,
        interactive=interactive,
        entrypoint=entrypoint,
        dry_run=dry_run,
        settings=settings.build
    )

    if isinstance(ret, int):
        raise typer.Exit(ret)



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from bmde.core.spec import BaseSpec
from bmde.core.types import BackendOptions


@dataclass
class SpecExecOpts(BaseSpec):
    backend: BackendOptions
    background: bool
    dry_run: bool
    interactive: bool
    entrypoint: Optional[Path]
    arguments: Optional[list[str]]



from pathlib import Path
from typing import Optional, List

from pydantic import BaseModel

from bmde.core.types import BackendOptions


class ExecutionSettings(BaseModel):
    entrypoint: Optional[Path] = None
    arguments: Optional[List[str]] = None
    background: Optional[bool] = False
    dry_run: Optional[bool] = False
    interactive: bool = True
    backend: Optional[BackendOptions] = None



from typing import Optional

from pydantic import BaseModel

from bmde.commands.run.settings import RunSettings
from bmde.config.command_settings import ExecutionSettings
from bmde.core.types import DockerOutputOptions


class DebugSettings(BaseModel):
    run: RunSettings = RunSettings()
    execution_settings: ExecutionSettings = ExecutionSettings()

    docker_screen: Optional[DockerOutputOptions] = DockerOutputOptions.HOST
    docker_network: Optional[str] = None



from __future__ import annotations

import os
from pathlib import Path
from subprocess import Popen
from typing import Optional

from bmde.core import logging
from bmde.core.exec import ExecOptions
from .service import BuildService
from .settings import BuildSettings
from .spec import BuildSpec, BuildSpecOpts
from ...config.loader import load_settings
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import BackendOptions

log = logging.get_logger(__name__)


def create_build_spec(
    d: Optional[Path],
    arguments: Optional[list[str]] = None,
    backend: Optional[BackendOptions] = None,
    background: bool = False,
    interactive: bool = True,
    dry_run: bool = False,
    entrypoint: Optional[Path] = None,
    settings: Optional[BuildSettings] = None,
) -> BuildSpec:
    """
    Creates the BuildSpec from the provided arguments.
    """
    if d is None:
        d = Path(os.getcwd())

    if settings is None:
        settings = BuildSettings()

    # Default args are prepended to CLI args
    passed_args: list[str] = []
    if arguments is not None:
        passed_args += arguments
    if settings.execution_settings.arguments is not None:
        passed_args += settings.execution_settings.arguments

    return BuildSpec(
        BuildSpecOpts=BuildSpecOpts(
            d=d,
        ),
        SpecExecOpts=SpecExecOpts(
            backend=backend if backend is not None else (settings.execution_settings.backend if settings.execution_settings.backend is not None else BackendOptions.DOCKER),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=passed_args,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                background
                if background is not None
                else settings.execution_settings.background
            ),
            interactive=(
                interactive
                if interactive is not None
                else settings.execution_settings.interactive
            ),
        ),
    )


def execute_build(spec: BuildSpec) -> int | Popen[bytes]:
    """
    Executes the BuildSpec.
    """
    return BuildService().run(
        spec.BuildSpecOpts,
        ExecOptions(
            arguments=spec.SpecExecOpts.arguments,
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            interactive=spec.SpecExecOpts.interactive,
            backend=spec.SpecExecOpts.backend,
            entrypoint=spec.SpecExecOpts.entrypoint
        ),
    )


def build_command(
    d: Optional[Path],
    arguments: Optional[list[str]] = None,
    backend: Optional[BackendOptions] = None,
    background: bool = False,
    interactive: bool = True,
    dry_run: bool = False,
    entrypoint: Optional[Path] = None,
    settings: Optional[BuildSettings] = None,
) -> int | Popen[bytes]:
    """
    CLI Entrypoint. Orchestrates spec creation and execution.
    """
    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(
            settings=full_settings
        )
        settings = full_settings.build

    spec = create_build_spec(
        d=d,
        arguments=arguments,
        backend=backend,
        background=background,
        interactive=interactive,
        dry_run=dry_run,
        entrypoint=entrypoint,
        settings=settings,
    )

    return execute_build(spec)



import typer
from typing import Any, List, Tuple, Optional

from bmde.config.loader import load_settings
from bmde.core import logging
from bmde.core.logging import configure_logging_from_settings
from bmde.core.service import Service
from bmde.core.types import BackendOptions
from bmde.config.schema import Settings
from bmde.commands.git.service import GitService
from bmde.commands.run.service import RunService
from bmde.commands.build.service import BuildService
from bmde.commands.debug.service import DebugService
from bmde.commands.patch.service import PatchService

log = logging.get_logger(__name__)


def check_command(settings: Settings) -> None:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(
            settings=full_settings,
            obfuscate_sensibles=full_settings.logging.hide_sensibles,
        )

    services: List[Tuple[str, Service[Any, Any], Optional[BackendOptions]]] = [
        ("Git", GitService(), settings.git.execution_settings.backend),
        ("Run", RunService(), settings.run.execution_settings.backend),
        ("Build", BuildService(), settings.build.execution_settings.backend),
        ("Debug", DebugService(), settings.debug.execution_settings.backend),
        ("Patch", PatchService(), settings.patch.execution_settings.backend),
    ]

    msg = "\nSummary of backend availability for the BMDE services in this machine (🚫 means not available, ✅ means available):\n"
    for name, service, selected_backend in services:
        msg += f"\n  * Service: {name}\n"

        if selected_backend:
            backend_impl = service.mapping.get(selected_backend)
            if backend_impl:
                available = backend_impl.is_available()
                icon = "✅" if available else "🚫"
                msg += f"    Selected backend: {selected_backend.value} {icon}\n"
            else:
                msg += f"    Selected backend: {selected_backend.value} (Unknown backend?)\n"
        else:
            msg += "    Selected backend: Auto (None configured)\n"

        msg += "    Other backends for this service (by order of preference):\n"
        for backend_opt in service.order:
            if selected_backend and backend_opt == selected_backend:
                continue

            backend_impl = service.mapping.get(backend_opt)
            if backend_impl:
                available = backend_impl.is_available()
                icon = "✅" if available else "🚫"
                msg += f"      - {backend_opt.value}: {icon}\n"

    typer.echo(msg)



"""
Defines the schema of settings of the application
"""

from pydantic import BaseModel

from bmde.config.logging_settings import LoggingSettings
from bmde.commands.build.settings import BuildSettings
from bmde.commands.run.settings import RunSettings
from bmde.commands.git.settings import GitSettings
from bmde.commands.patch.settings import PatchSettings
from bmde.commands.debug.settings import DebugSettings


class Settings(BaseModel):
    logging: LoggingSettings = LoggingSettings()
    run: RunSettings = RunSettings()
    build: BuildSettings = BuildSettings()
    git: GitSettings = GitSettings()
    patch: PatchSettings = PatchSettings()
    debug: DebugSettings = DebugSettings()



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from bmde.core.spec import BaseSpec
from bmde.core.spec_opts import SpecExecOpts


@dataclass
class GitSpecOpts(BaseSpec):
    d: Path
    ssh_username: Optional[str]
    ssh_password: Optional[str]
    ssh_host: Optional[str]
    git_name: Optional[str]
    git_email: Optional[str]
    vpn_username: Optional[str]
    vpn_password: Optional[str]
    vpn_host: Optional[str]
    vpn_port: Optional[int]
    vpn_realm: Optional[str]
    vpn_cert: Optional[str]
    vpn_test_dns: Optional[str]
    vpn_test_ip: Optional[str]


@dataclass
class GitSpec(BaseSpec):
    SpecExecOpts: SpecExecOpts
    GitSpecOpts: GitSpecOpts



from typing import Optional

from pydantic import BaseModel

from bmde.config.command_settings import ExecutionSettings


class VpnAuthSettings(BaseModel):
    enabled: bool = True
    username: Optional[str] = None
    password: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    cert: Optional[str] = None
    realm: Optional[str] = None
    test_dns: Optional[str] = None
    test_ip: Optional[str] = None


class GitSshSettings(BaseModel):
    username: Optional[str] = None
    password: Optional[str] = None
    host: Optional[str] = None


class GitConfigSettings(BaseModel):
    name: Optional[str] = None
    email: Optional[str] = None


class GitSettings(BaseModel):
    execution_settings: ExecutionSettings = ExecutionSettings()

    git: GitConfigSettings = GitConfigSettings()
    ssh: GitSshSettings = GitSshSettings()
    vpn: VpnAuthSettings = VpnAuthSettings()



from __future__ import annotations

from pathlib import Path
from subprocess import Popen
from typing import Optional

from bmde.core import logging
from bmde.core.exec import ExecOptions
from .service import PatchService
from .settings import PatchSettings
from .spec import PatchSpec, PatchSpecOpts
from ...config.loader import load_settings
from ...core.file_utils import resolve_nds
from ...core.logging import configure_logging_from_settings
from ...core.spec_opts import SpecExecOpts
from ...core.types import BackendOptions

log = logging.get_logger(__name__)


def create_patch_spec(
    d: Optional[Path],
    nds_rom: Optional[Path] = None,
    arguments: Optional[list[str]] = None,
    backend: Optional[BackendOptions] = None,
    background: bool = False,
    dry_run: bool = False,
    entrypoint: Optional[Path] = None,
    interactive: bool = True,
    settings: Optional[PatchSettings] = None,
) -> PatchSpec:

    if d is None:
        d = Path.cwd()

    nds_resolved, _ = resolve_nds(nds_rom, cwd=d)

    if settings is None:
        settings = PatchSettings()

    return PatchSpec(
        PatchSpecOpts=PatchSpecOpts(nds_rom=nds_resolved),
        SpecExecOpts=SpecExecOpts(
            backend=(
                backend
                if backend is not None
                else (
                    settings.execution_settings.backend
                    if settings.execution_settings.backend is not None
                    else BackendOptions.DOCKER
                )
            ),
            entrypoint=(
                entrypoint
                if entrypoint is not None
                else settings.execution_settings.entrypoint
            ),
            arguments=arguments,
            dry_run=(
                dry_run if dry_run is not None else settings.execution_settings.dry_run
            ),
            background=(
                background
                if background is not None
                else settings.execution_settings.background
            ),
            interactive=interactive,
        ),
    )


def execute_patch(spec: PatchSpec) -> int | Popen[bytes]:
    return PatchService().run(
        spec.PatchSpecOpts,
        ExecOptions(
            dry_run=spec.SpecExecOpts.dry_run,
            background=spec.SpecExecOpts.background,
            entrypoint=spec.SpecExecOpts.entrypoint,
            arguments=spec.SpecExecOpts.arguments,
            backend=spec.SpecExecOpts.backend,
            interactive=spec.SpecExecOpts.interactive,
        ),
    )


def patch_command(
    d: Optional[Path],
    nds_rom: Optional[Path] = None,
    arguments: Optional[list[str]] = None,
    backend: Optional[BackendOptions] = None,
    background: bool = False,
    dry_run: bool = False,
    entrypoint: Optional[Path] = None,
    interactive: bool = True,
    settings: Optional[PatchSettings] = None,
) -> int | Popen[bytes]:

    if settings is None:
        full_settings = load_settings()
        configure_logging_from_settings(settings=full_settings)
        settings = full_settings.patch

    spec = create_patch_spec(
        d=d,
        nds_rom=nds_rom,
        arguments=arguments,
        backend=backend,
        background=background,
        dry_run=dry_run,
        entrypoint=entrypoint,
        interactive=interactive,
        settings=settings,
    )

    return execute_patch(spec)
